#include "stacktool.h"

static const int DEFAULT_QUEUE_SIZE = 100000;

static branch_idx_t add_edge (branch_list *list,
			      code_addr_t dest,
			      bool dead,
                              cost_t cost)
{
  branch_idx_t slot;

  assert (list != NULL);
  slot = (branch_idx_t)list->len;
  list->len++;

  list->list = (struct branch_elt *) xrealloc (list->list, 
					       list->len * sizeof (struct branch_elt));

  list->list[slot].target = abs_loc(dest,NO_CTX);
  list->list[slot].dead = dead;
#ifdef WCET_ANALYSIS
  list->list[slot].cost = cost;
#endif

  return slot;
}

static branch_idx_t add_in_out (code_addr_t src, code_addr_t dest, bool dead, cost_t cost)
{
  branch_idx_t forward_index;
  code_ctxt_t i;
  branch_idx_t tmp;

  decode_addr (dest);

  if (debug_control) printf ("adding forward edge from %x to %x\n", src, dest);

  forward_index = add_edge (&all_code[src]->control.edges_out, dest, dead,cost);
  for (i = (code_ctxt_t)0; i < (code_ctxt_t)all_code[src]->state_cnt; i++) { 
    tmp = add_edge (&all_code[src]->states[i].out_edges, dest, dead,cost);
    assert (tmp == forward_index);
  }

  if (debug_control) printf ("adding backward edge from %x to %x\n", dest, src);

  branch_idx_t tmp2 = add_edge (&all_code[dest]->control.edges_in, src, false, cost);
  for (i = (code_ctxt_t)0; i < (code_ctxt_t)all_code[dest]->state_cnt; i++) {
    tmp = add_edge (&all_code[dest]->states[i].in_edges, src, false, cost);
    assert (tmp == tmp2);
  }

  return forward_index;
}

enum control_cmd {
  CTL_NO_CMD = 1699,
  CTL_FORWARD,
  CTL_NEW_FLOW,
  CTL_NEW_ICALL,
  CTL_NEW_TOS_TASK,
  CTL_NOP
};

struct control_queue_elt {
  enum control_cmd cmd;
  code_addr_t addr;
};

static struct control_queue_elt control_queue[DEFAULT_QUEUE_SIZE];
static int control_queue_index = 0;

static const char *cmd_name (enum control_cmd cmd)
{
  switch (cmd) {
  case CTL_NO_CMD:
    return "no command";
  case CTL_FORWARD:
    return "forward";
  case CTL_NEW_FLOW:
    return "new flow";
  case CTL_NEW_TOS_TASK:
    return "new tinyos task";
  case CTL_NOP:
    return "nop";
  default:
    die ();
  }
}

static void enqueue_control_bare (int idx,
				  enum control_cmd cmd, 
				  code_addr_t addr)
{
  if (debug_control) {
    if (cmd != CTL_NOP) {
      printf ("enqueue_control [%d]: cmd [%s] addr %x\n",
	      control_queue_index, cmd_name (cmd), addr);
    }
  }

  control_queue[idx].cmd = cmd;
  control_queue[idx].addr = addr;
}

static void enqueue_control (enum control_cmd cmd, 
			     code_addr_t addr)
{
  enqueue_control_bare (control_queue_index, cmd, addr);
  control_queue_index++;
  assert (control_queue_index < DEFAULT_QUEUE_SIZE);
}

void enqueue_task (code_addr_t addr)
{
  enqueue_control (CTL_NEW_TOS_TASK, addr);
}

static void dequeue_control (enum control_cmd *cmd, 
			     code_addr_t *addr)
{
  if (control_queue_index == 0) {
    *cmd = CTL_NO_CMD;
    return;
  }
  control_queue_index--;
  *cmd = control_queue[control_queue_index].cmd;
  *addr = control_queue[control_queue_index].addr;

  if (debug_control && *cmd != CTL_NOP) {
    printf ("\ndequeue_control [%d]: cmd [%s] addr %x\n",
	    control_queue_index, 
	    cmd_name (*cmd), 
	    (*addr));
  }
}

branch_idx_t add_branch (code_addr_t src, 
			 code_addr_t dest, 
			 bool dead, 
			 cost_t cost) 
{
  branch_idx_t forward_index;

  assert (src != NO_CODE_ADDR);
  assert (dest != NO_CODE_ADDR);

  forward_index = add_in_out (src, dest, dead, cost);
  enqueue_control (CTL_FORWARD, dest);

  return forward_index;
}

static void control_effect (code_addr_t addr)
{
  assert (addr != NO_CODE_ADDR);

  decode_addr (addr);
  assert (all_code[addr] != NULL);
  enum AVR_OPCODE opcode = all_code[addr]->decode.opcode;

  // don't revisit
  if (all_code[addr]->touched) goto end;

  if (debug_control) {
    printf ("addr = %x\n  touched = %s\n", 
	    addr, 
	    bool_to_str(all_code[addr]->touched));
    printf ("  insn = %s\n", 
	    opcode_to_string (all_code[addr]->decode.opcode));
  }

  all_code[addr]->touched = true;
  
  switch (opcode) {

  case AVR_SET_STACK_PTR:
  case AVR_DEC_STACK_PTR:
  case AVR_INC_STACK_PTR:
    {
      assert (!all_code[addr]->decode.target_set);
      cost_t cost = all_code[addr]->decode.size; // ToDo: hacky guess
      (void) add_branch (addr, nextPC(addr), false, cost);
      break;
    }

  case AVR_breq:
  case AVR_brge:
  case AVR_brne:
  case AVR_brcs:
  case AVR_brcc:
  case AVR_brlt:
  case AVR_brhc:
  case AVR_brhs:
  case AVR_brid:
  case AVR_brie:
  case AVR_brmi:
  case AVR_brpl:
  case AVR_brtc:
  case AVR_brts:
  case AVR_brvc:
  case AVR_brvs:
  case AVR_brbc:
  case AVR_brbs:
    assert (all_code[addr]->decode.target_set);
    all_code[addr]->control.index_taken = add_branch (addr, 
						      all_code[addr]->decode.target, 
						      true,
                                                      (cost_t)2);
    all_code[addr]->control.index_not_taken = add_branch (addr, 
							  nextPC(addr), 
							  true,
                                                          (cost_t)1);
    break;

  case AVR_jmp:
  case AVR_rjmp:
    assert (all_code[addr]->decode.target_set);
    if (all_code[addr]->decode.target == 0) {
#if VERBOSE_CTRL_FLOW
      if (!quiet) printf ("ignoring jump to 0x0\n");
#endif
    } else {
      (void) add_branch (addr, all_code[addr]->decode.target, false, 
                         (cost_t)(opcode==AVR_jmp?3:2));
    }
    break;

  case AVR_ijmp:
  case AVR_eijmp:
    // do nothing here -- we'll add control flow edges after doing
    // some dataflow analysis
    break;

  case AVR_call:
  case AVR_rcall:
    {
      // avoid calling address 0 -- a few kernels use this to reset the processor!
      if (all_code[addr]->decode.target == 0) {
	printf ("Ignoring a call to address 0x0;\n");
	printf ("this is sometimes an idiom for resetting the CPU.\n");
      } else {
	(void) add_branch (addr, all_code[addr]->decode.target, false, 
			   (cost_t)(opcode==AVR_call?5:4));
	enqueue_control (CTL_FORWARD, nextPC(addr));
      }
    }
    break;
      
  case AVR_icall:
  case AVR_eicall:
    // find the branch target later!
    enqueue_control (CTL_FORWARD, nextPC(addr));
    break;

    // no new control flow
  case AVR_ret:
    // find return branches later
  case AVR_reti:
    break;

  case AVR_sbrs:
  case AVR_sbrc:
  case AVR_cpse:
  case AVR_sbic:
  case AVR_sbis:
    {
      // skip size depends on structure of next insn
      code_addr_t next = nextPC(addr);
      decode_addr (next);
      assert (all_code[next] != NULL);
      all_code[addr]->control.index_taken = 
	add_branch (addr, nextPC(next), true, 1+all_code[next]->decode.size/2);
      all_code[addr]->control.index_not_taken = 
	add_branch (addr, next, true, 1);
    }
    break;

  case AVR_out:
    assert (all_code[addr]->decode.ioaddr_set);
    /*lint -fallthrough */
  case AVR_adc:
  case AVR_add:
  case AVR_adiw:
  case AVR_and:
  case AVR_andi:
  case AVR_asr:
  case AVR_bld:
  case AVR_break:
  case AVR_bst:
  case AVR_cbi:
  case AVR_clh:
  case AVR_cli:
  case AVR_cln:
  case AVR_cls:
  case AVR_clt:
  case AVR_clv:
  case AVR_clz:
  case AVR_com:
  case AVR_cp:
  case AVR_cpc:
  case AVR_cpi:
  case AVR_dec:
  case AVR_elpm1:
  case AVR_elpm2:
  case AVR_elpm3:
  case AVR_eor:
  case AVR_fmul:
  case AVR_fmuls:
  case AVR_fmulsu:
  case AVR_in:
  case AVR_inc:
  case AVR_ldi:
  case AVR_lpm1:
  case AVR_lpm2:
  case AVR_lpm3:
  case AVR_lsr:
  case AVR_mov:
  case AVR_movw:
  case AVR_mul:
  case AVR_muls:
  case AVR_mulsu:
  case AVR_neg:
  case AVR_nop:
  case AVR_or:
  case AVR_ori:
  case AVR_pop:
  case AVR_push:
  case AVR_ror:
  case AVR_sbc:
  case AVR_sbci:
  case AVR_sbi:
  case AVR_sbiw:
  case AVR_seh:
  case AVR_sei:
  case AVR_sen:
  case AVR_ses:
  case AVR_set:
  case AVR_sev:
  case AVR_sez:
  case AVR_sleep:
  case AVR_spm:
  case AVR_sub:
  case AVR_subi:
  case AVR_swap:
  case AVR_wdr:
  case AVR_ld1x:
  case AVR_ld2x:
  case AVR_ld3x:
  case AVR_ld1y:
  case AVR_ld2y:
  case AVR_ld3y:
  case AVR_ldd1y:
  case AVR_ld1z:
  case AVR_ld2z:
  case AVR_ld3z:
  case AVR_ldd1z:
  case AVR_sec:
  case AVR_st1x:
  case AVR_st2x:
  case AVR_st3x:
  case AVR_st1y:
  case AVR_st2y:
  case AVR_st3y:
  case AVR_std1y:
  case AVR_st1z:
  case AVR_st2z:
  case AVR_st3z:
  case AVR_std1z:
    assert (!all_code[addr]->decode.target_set);
    assert (all_code[addr]->decode.size == 2);
    (void) add_branch (addr, nextPC(addr), false, avr_cost(opcode));
    break;

  case AVR_lds:
  case AVR_sts:
    assert (!all_code[addr]->decode.target_set);
    assert (all_code[addr]->decode.size == 4);
    (void) add_branch (addr, nextPC(addr), false, avr_cost(opcode));
    break;

  default:
    printf ("oops: unknown opcode %x at addr %x\n", opcode, addr);
    die ();
  }

 end:
  ;
}

void control_flow_setup (void)
{
  if (!quiet && debug_control > 0) printf ("origin is %x\n", origin);

  all_code = (avr_insn **) xmalloc (code_size * sizeof (avr_insn *));
  sp_consistent = (bool *) xmalloc (code_size * sizeof (bool));
	 
  for (unsigned int i=0; i<code_size; i++) {
    all_code[i] = NULL;
    sp_consistent[i] = true;
  }

  if (debug_control) {
    printf ("allocated all_code for %d instructions\n", code_size);
  }

  for (unsigned int i=0; i<platform->num_vectors; i++) {
    enqueue_control (CTL_NEW_FLOW, platform->int_vecs[i].start_addr);
  }

  if (true) {
    unsigned int i=0;
    while (true) {
      code_addr_t addr = origin + (VECTOR_SIZE * i);
      decode_addr (addr);
      if (all_code[addr] != NULL && 
	  all_code[addr]->decode.opcode == AVR_jmp) {
	undecode_addr (addr); // avoid side effects
	i++;
      } else {
	undecode_addr (addr); // avoid side effects
	break;
      }
    }

    if (i != platform->num_vectors) {
      printf ("oops -- didn't find %d vectors as expected!\n",
	      platform->num_vectors);
      die ();
    }
  }
}

bool control_flow_step (void)
{
  code_addr_t addr;
  enum control_cmd cmd;

  dequeue_control (&cmd, &addr);

  switch (cmd) {

  case CTL_NO_CMD:
    return false;

  case CTL_NOP:
    break;

  case CTL_NEW_FLOW: 
    control_effect (addr);
    if (debug_control) printf ("enqueuing dataflow for new flow\n");
    enqueue_data (abs_loc(NO_CODE_ADDR, (code_ctxt_t)0),  // prev
		  abs_loc(addr, (code_ctxt_t)0),          // cur
		  abs_loc(NO_CODE_ADDR, (code_ctxt_t)0),  // ret
		  abs_loc(NO_CODE_ADDR, (code_ctxt_t)0)); // head
    break;

  case CTL_NEW_ICALL:
    break;
    
  case CTL_NEW_TOS_TASK:   
    code_addr_t my_ret_addr;
    code_addr_t my_src_addr;
    branch_idx_t idx;
    
    control_effect (addr);
    
    if (num_schedulers == 0) {
      printf ("WARNING: ignoring tinyos task because we don't\n");
      printf ("know any schedulers (this is probably bad)\n");
      die ();
    }
    
    for(int i=0; i<num_schedulers; ++i) {
      class scheduler* call = &schedulers[i];
      
      my_src_addr = call->addr;
      my_ret_addr = nextPC(call->addr);
      
      if (debug_control) printf ("enqueuing dataflow for new icall\n");
      
      if (debug_control) printf ("found an icall to 0x%x %s\n",
				 addr, all_code[addr]->decode.label);
      
      idx = add_in_out (my_src_addr, addr, false, (cost_t)4);
      
      add_call_info (abs_loc(addr,        (code_ctxt_t)0), // target
		     abs_loc(my_src_addr, (code_ctxt_t)0), // current
		     abs_loc(my_ret_addr, (code_ctxt_t)0), // successor
		     call->stash_ret,
		     call->stash_head,
		     idx);  
      
      enqueue_data (abs_loc(my_src_addr, (code_ctxt_t)0),  // prev
		    abs_loc(addr, (code_ctxt_t)0),         // cur
		    abs_loc(my_ret_addr, (code_ctxt_t)0),  // ret
		    abs_loc(addr, (code_ctxt_t)0));        // head -- no head, we're returning to main
    }
    break;

  case CTL_FORWARD:
    control_effect (addr);
    break;

  default:
    die ();
  }

  return true;
}
