#include "stacktool.h"

const mem_loc_t mem_loc_t::bottom = mem_loc_t(0,0);
const mem_loc_t mem_loc_t::zero = mem_loc_t(0xff,0);

ostream& operator << (ostream& os, const srcloc &loc) {
  return os << loc.filename << "::" << dec << loc.line_number << hex;
};

ostream& operator << (ostream& os, const enum AVR_OPCODE b) {
  return os << opcode_to_string(b);
};


ostream& operator << (ostream& os, const class mem_loc_t &b) {
  return os << "known = " << b.known << ", val = " << b.val;
};

ostream& operator << (ostream& os, const class abs_loc &b) {
  os << hex << b.addr << '.';
  if (b.ctx != NO_CTX) 
    return os << dec << b.ctx << hex;
  else 
    return os << "?";
};

// Just a standard print function except that it takes an extra argument
// to let us terminate recursion.
ostream& print_context (ostream& os, abs_loc_t b, 
                        set<abs_loc_t,1000> &visited) {
#if 0
  // a very minimalist version that produces output just like we used to
  // produce
  return os << b;
#else
  // a more full-featured version that traces down call chains
  // Guard against infinite loops
  static int enter=0;
  enter++;
  if (enter > 100) {
    os << b;
    die();
  }

  if (visited.member(b)) {
    // Square brackets indicate recursion.
    os << "[" << b << "]";
    goto retrn;
  }
  visited.insert(b);
  assert( !visited.is_bottom() ); // means we need to use a bigger set

  assert(all_code[b.addr]);
  if (b.ctx < all_code[b.addr]->state_cnt && state(b)) {
    // Find the head of the current function
    // Warning: if an instruction has multiple heads, we only get
    // one of them.
    abs_loc_t head = state(b)->head;
    if (head.addr == NO_CODE_ADDR) goto retrn;

    os << head.addr << "/";
  
    // Chase down all the return addresses
    if (state(head)->num_calls != 1) {
      os << "{";
    }
    for(int i = 0; i != state(head)->num_calls; ++i) {
      if (0 != i) { os << ","; }
      os << state(head)->calls[i].ret << ":";
      print_context(os,state(head)->calls[i].ret,visited);
    }
    if (state(head)->num_calls != 1) {
      os << "}";
    }
  } else {
    // Dead instruction.  For example, it could be an
    // unreachable return address.
    os << "<" << b << ">";
  }
 retrn:
  --enter;
  return os;
#endif
};

ostream& print_context (ostream& os, abs_loc_t b) {
  set<abs_loc_t,1000> visited;
  return print_context(os,b,visited);
};

static void print_int_bit (const abs_loc_t loc,
			   FILE *outf)
{
  if (is_status_bit_known (&state(loc)->mem, I_BIT)) {
    if (is_status_bit_set (&state(loc)->mem, I_BIT)) {
      fprintf (outf, "+");
    } else {
      fprintf (outf, "-");
    }
  } else {
    fprintf (outf, "?");
  }
}

void print_int_bits (const abs_loc_t loc,
		     FILE *outf)
{
  for (unsigned int i=1; i<platform->num_vectors; i++) {
    if (int_known (loc, i)) {
      if (int_enabled (loc, i)) {
	fprintf (outf, "+");
      } else {
	fprintf (outf, "-");
      }
    } else {
      fprintf (outf, "?");
    }
  }
}

/* 
 * could optimize this by rearranging the data structures and then
 * working on a 32-bit word at a time; probably not worth it
 */
static void merge_state (const memory *oldl, 
			 memory *newl)
{
  for (data_addr_t addr=0; addr < (data_addr_t) platform->state_size; addr++) {
    newl->core[addr] = merge (oldl->core[addr], newl->core[addr]);
    newl->modified[addr] |= oldl->modified[addr];
  }

#if TRACK_STACK
  if (oldl->stack.size != newl->stack.size) {
    // FIXME -- why does this happen?
    newl->stack.set_to_bottom();
  } else if (oldl->stack.is_bottom() || newl->stack.is_bottom()) {
    if (!oldl->stack.is_bottom() || !newl->stack.is_bottom()) {
      printf ("hrm: merging a bottom-stack with a non-bottom stack\n");
    }
    newl->stack.set_to_bottom();
  } else {
    for(int i=0; i<newl->stack.size; i++) {
      newl->stack.elts[i] = merge (oldl->stack.elts[i], newl->stack.elts[i]);
    }
  }
#endif

}

int count_bits(data_val_t x)
{
  int r = 0;
  while (x != 0) {
    if (x & 1) r++;
    x >>= 1;
  }
  return r;
}

long long int count_bits (const memory *s)
{
  long long int r = 0;
  data_addr_t addr;
  for (addr=0; addr < (data_addr_t) platform->regs.len; addr++) {
    r += count_bits(s->core[addr].known);
  }
  r += count_bits(s->core[io_to_mem(SREG_ADDR)].known);
  return r;
}

void print_brief_dataflow (code_addr_t addr, FILE *outf)
{
  code_ctxt_t i;
  bool state_init = false;
  state_inst tmp_state;
  struct avr_insn_t *insn = all_code[addr];

  // fprintf (outf, "[%d] ", insn->state_cnt);

  if (insn->state_cnt == 0) return;

  assert (insn->state_cnt > 0);

  for (i = (code_ctxt_t)0; i < (code_ctxt_t)insn->state_cnt; i++) {

    const abs_loc_t loc = abs_loc_t (addr, i);

    if (insn->states[i].visit_cnt == 0) continue;

    if (i == (code_ctxt_t)0) {
      tmp_state = insn->states[i];
      state_init = true;
    } else {
      merge_state (&insn->states[i].mem, &tmp_state.mem);		 
    }

    print_int_bit (loc, outf);
  }

  fprintf (outf, " ");

  // if (state_init) print_int_bit (&tmp_state, outf);
}

static void print_prefix (code_addr_t addr, 
			  data_addr_t reg_addr, 
			  FILE *outf, ostream &os)
{
  if (reg_addr < (data_addr_t) platform->regs.len) {
    fprintf (outf, "  r%02d:  ", reg_addr);
  } else {
    if (reg_addr == io_to_mem (SREG_ADDR)) {
      os << "  SREG: ";
    } else {
      fprintf (outf, "  i/o %02x: ", mem_to_io (reg_addr));
    }
  }
  os << "addr = " << hex << addr << ", ";
}
  
static void print_state (abs_loc_t loc1, 
			 data_addr_t addr2, 
			 FILE *outf, ostream &os)
{
  assert (state(loc1)->visit_cnt > 0);

  // don't print what we don't know
#if 1
  if (memry(loc1,addr2)->known == 0x0 &&
      addr2 != io_to_mem (SREG_ADDR)) return;
#endif  

  print_prefix (loc1.addr, addr2, outf, os);

  // mask off unknown bits while printing 
  { 
    data_kno_t k = memry(loc1,addr2)->known;
    data_val_t v = memry(loc1,addr2)->val;

    fprintf (outf, "known = %02x, val = %02x\n", k, v&k);
  }
}

static int SP_val (abs_loc_t *loc)
{
  return
    (ioport (*loc, SPH_ADDR)->val << 8) +
    ioport (*loc, SPL_ADDR)->val;
}

static void print_SP (abs_loc_t loc, FILE *outf)
{
  if (ioport (loc, SPH_ADDR)->known != 0xff ||
      ioport (loc, SPL_ADDR)->known != 0xff) {
    return;
  }

  fprintf (outf, "  SP = %x\n", SP_val (&loc));
}

static void print_verbose_dataflow_ctx (abs_loc_t loc, FILE *outf, ostream &os)
{
  if (state(loc)->visit_cnt > 0) {
    
    fprintf (outf, "\nmachine state %x context %d\n", 
	     loc.addr, loc.ctx);
    
    if (debug_data > 0) {
      for (data_addr_t x=0; x < (data_addr_t) platform->regs.len; x++) {
        print_state (loc, x, outf, os);
      }
      for(io_data_addr_t i=0; i < platform->io.len; ++i) {
        print_state (loc, io_to_mem (i), outf, os);
      }
      print_state (loc, io_to_mem (SREG_ADDR), outf, os);
      os << "  Stack:   ";
      print_context(os, loc) << endl;
      print_state (loc, io_to_mem (SPH_ADDR), outf, os);
      print_state (loc, io_to_mem (SPL_ADDR), outf, os);
      print_SP (loc, outf);
      fprintf (outf, "  stack top = %x\n", state(loc)->mem.stack_top);
      fprintf (outf, "  stack ptr = %x\n", state(loc)->mem.shadow_sp);
    }
#if WCET_ANALYSIS
    if (debug_wcet > 0) {
      if (debug_wcet > 1) {
        for(int i=0; i<state(loc)->num_ret_targets; ++i) {
          os << "  ReturnTo[" << dec << i << hex << "]: " << state(loc)->ret_targets[i] << endl;
        }
        for(unsigned int i=0; i<state(loc)->out_edges.len; ++i) {
          struct branch_elt *b = &state(loc)->out_edges.list[i];
          if (!b->dead) {
            os << "  Next[" << dec << i << hex << "]: " << b->target << endl;
          }
        }
        for(unsigned int i=0; i<state(loc)->in_edges.len; ++i) {
          struct branch_elt *b = &state(loc)->in_edges.list[i];
          if (!b->dead) {
            os << "  Prev[" << dec << i << hex << "]: " << b->target << endl;
          }
        }
        for(int i=0; i<state(loc)->num_rets; ++i) {
          struct ret_addr_ctx *b = &state(loc)->rets[i];
          os << "  Ret[" << dec << i << hex << "]: " << b->ret << endl;
        }
        for(int i=0; i<state(loc)->num_calls; ++i) {
          struct call_addr_ctx *b = &state(loc)->calls[i];
          os << "  Call[" << dec << i << hex << "]: " 
             << b->call << "/" << b->ret
             << endl;
        }
        os << "  Terminate: " << state(loc)->terminate_analysis << endl;
      }
      if (all_code[loc.addr]->exn_count >= 0) {
        os << "  Execution count: " << dec << all_code[loc.addr]->exn_count << hex << endl;
      }
      if (state(loc)->wcet >= 0) {
        os << "  WCET: " << dec << state(loc)->wcet << hex << endl;
      }
    }
#endif

    os << endl;
    
  } else {
    os << "\n(not init)\n";      
  }
}

void print_verbose_dataflow (code_addr_t addr, FILE *outf, ostream &os)
{
  queue<srcloc> *locs = 0;
  if (addr_to_srclocs(addr,&locs)) {
    for(queue<srcloc>::iterator i = locs->begin(); i != locs->end(); ++i) {
      os << "  " << *i << endl;
    }
  }
  for (unsigned i=0; i<all_code[addr]->state_cnt; i++) {
    print_verbose_dataflow_ctx (abs_loc(addr, (code_ctxt_t)i), outf, os);
  }
}

static void copy_branch_list (const branch_list *src, 
			      branch_list *dst,
                              code_ctxt_t ctx)
{
  assert (src != NULL);
  assert (dst != NULL);
  assert (dst->list == NULL);

  if (src->len > 0) {
    assert (src->list != NULL);
    dst->list = (struct branch_elt *) 
      xmalloc ((size_t)src->len * sizeof (struct branch_elt));
  } else {
    assert (src->list == NULL);
  }

  dst->len = src->len;

  for (unsigned i=0; i < src->len; i++) {
    dst->list[i] = src->list[i];
    if (dst->list[i].target.ctx == NO_CTX) {
      dst->list[i].target.ctx = ctx;
    }
  }
}

/*
 * dynamically allocate a new machine state at a particular program
 * address, copying control-flow edges from the "master copy"
 */
code_ctxt_t create_new_context (code_addr_t addr)
{
  code_ctxt_t slot = (code_ctxt_t) all_code[addr]->state_cnt;
  all_code[addr]->state_cnt++;

  if (debug_data > 1) {
    cout << "  create_new_context " << slot << " at " << addr;
    cout << " with " << all_code[addr]->control.edges_out.len
         << " out branches to ";
	    
    for (unsigned int i=0; i<all_code[addr]->control.edges_out.len; i++) {
      printf ("%x ", 
	      all_code[addr]->control.edges_out.list[i].target.addr);
    }
    cout << endl;
  }
    
  all_code[addr]->states = (state_inst *) 
    xrealloc (all_code[addr]->states, 
	      (size_t) all_code[addr]->state_cnt * 
	      sizeof (state_inst));

  all_code[addr]->states[slot].init();

  copy_branch_list (&all_code[addr]->control.edges_in, 
		    &all_code[addr]->states[slot].in_edges,
                    slot);
  copy_branch_list (&all_code[addr]->control.edges_out, 
		    &all_code[addr]->states[slot].out_edges,
                    slot);

  return slot;
}

void lazy_create_contexts (abs_loc_t ctx)
{
  assert( ctx.ctx != NO_CTX );
  decode_addr(ctx.addr);
  assert (all_code[ctx.addr] != NULL);
  while (ctx.ctx >= (code_ctxt_t) all_code[ctx.addr]->state_cnt) {
    (void) create_new_context (ctx.addr);
  }
}

struct data_queue_elt {
  abs_loc_t cur;
  abs_loc_t prev;
  abs_loc_t call;
  abs_loc_t head;
};

static const int DEFAULT_QUEUE_SIZE = 100000;

static struct data_queue_elt data_queue[DEFAULT_QUEUE_SIZE];
static int data_queue_index = 0;

void enqueue_data (abs_loc_t prev,
		   abs_loc_t cur,
		   abs_loc_t call,
		   abs_loc_t head)
{
  assert (cur.addr < (code_addr_t) code_size);

  if (debug_data > 1) {
    cout << "  enqueue_data("
         << "prev=" << prev 
         << ", cur=" << cur
         << ", call=" << call
         << ", head=" << head
         << ")" << endl;

    if (debug_data > 2) {
      for (code_ctxt_t i = (code_ctxt_t)0; i < (code_ctxt_t) all_code[cur.addr]->state_cnt; i++) {
	cout << "    context " << i 
	     << ": out edges = " << all_code[cur.addr]->states[i].out_edges.len
	     << endl;
      }
    }
  }

  assert (cur.ctx  != NO_CTX);
  assert (prev.ctx != NO_CTX);
  lazy_create_contexts (cur);

  assert (cur.ctx < (code_ctxt_t) all_code[cur.addr]->state_cnt);

  data_queue[data_queue_index].prev = prev;
  data_queue[data_queue_index].cur  = cur;
  data_queue[data_queue_index].call = call;
  data_queue[data_queue_index].head = head;

  data_queue_index++;
  assert (data_queue_index < DEFAULT_QUEUE_SIZE);
}

static void dequeue_data (abs_loc_t *prev, 
			  abs_loc_t *cur, 
			  abs_loc_t *call, 
			  abs_loc_t *head)
{
  if (data_queue_index == 0) {
    cur->addr = NO_CODE_ADDR;
    return;
  }
  data_queue_index--;

  *cur  = data_queue[data_queue_index].cur;
  *prev = data_queue[data_queue_index].prev;
  *call = data_queue[data_queue_index].call;
  *head = data_queue[data_queue_index].head;
}

static void print_stuff (abs_loc_t cur,
			 bool res,
			 const state_inst *copied_state)
{
  data_addr_t i;
  state_inst * const s = state(cur);

  printf ("  changed = %s, visit_cnt = %d, context = %d (range 0..%d), num out edges = %d\n", 
	  bool_to_str(res), 
	  s->visit_cnt, 
	  cur.ctx,
	  all_code[cur.addr]->state_cnt-1,
	  s->out_edges.len);
  
  if (!res || s->visit_cnt == 1) return;

  for (i=0; i < (data_addr_t) platform->state_size; i++) {
    if (copied_state->mem.core[i].known != s->mem.core[i].known ||
	((copied_state->mem.core[i].known & copied_state->mem.core[i].val) !=
	 (s->mem.core[i].known & 
	  s->mem.core[i].val))) {
      if (i < (data_addr_t) platform->regs.len) {
	printf ("    changed r%2d    : ", i);
      } else {
	printf ("    changed i/o %2x : ", mem_to_io(i));
      }
      printf ("old: known = %02x, val = %02x    ",
	      copied_state->mem.core[i].known,
	      copied_state->mem.core[i].val);
      printf ("new: known = %02x, val = %02x\n",
	      s->mem.core[i].known,
	      s->mem.core[i].val);
    }
  }
}

code_addr_t pair_to_addr (data_val_t rv1, data_val_t rv2)
{
  return (code_addr_t) (2*((rv2 << 8) + rv1));
}

static inline bool is_canonical_loc (const mem_loc m)
{
  return (m.val == (m.val & m.known));
}

static bool is_canonical_state (state_inst *st)
{
  data_addr_t addr2;

  for (addr2=0; addr2<MAX_STATE_SIZE; addr2++) {
    if (!is_canonical_loc (st->mem.core[addr2])) {
      if (debug_data > 1) {
	printf ("  %d\n", addr2);
      }
      return false;
    }
  }
  return true;
}

static bool dataflow_effect (abs_loc_t PC,
			     abs_loc_t prev )
{
  memory incoming_mem;

  // make a copy of the machine state so we can easily tell if
  // anything changes in order to see when a fixpoint is reached
  state_inst state_before_effect = *state(PC);

  // machine state is "incoming" from previous address (if any)
  if (prev.addr == NO_CODE_ADDR) {
    if (debug_data > 1) cout << "  incoming_mem = &initial_values" << endl;
    initialize_values(platform, incoming_mem.core);
#if TRACK_STACK
    incoming_mem.stack = vec<mem_loc_t>();
    incoming_mem.stack.init();
#endif
    incoming_mem.stack_top = INIT_SP;
    incoming_mem.shadow_sp = INIT_SP;
  } else {
    if (debug_data > 1) cout << "  incoming_mem = state(" << prev << ")" << endl;
    for(int i=0; i<MAX_STATE_SIZE; ++i) {
      incoming_mem.core[i] = state(prev)->mem.core[i];
    }
#if TRACK_STACK
    // ToDo: get rid of test on which instruction we're executing
    if (avr_is_call(prev)) {
      if (debug_data) cout << "Hack: using callee stack after a call" << endl;
      incoming_mem.stack = vec<mem_loc_t>();
      incoming_mem.stack.init();
    } else if (avr_is_ret(prev)) {
      if (debug_data) cout << "Hack: using caller stack after a return" << endl;
      abs_loc caller = prevPC(PC);
      assert(avr_is_call(caller));
      lazy_create_contexts(caller);
      incoming_mem.stack = state(caller)->mem.stack;
    } else {
      incoming_mem.stack = state(prev)->mem.stack;
    }
#endif
    incoming_mem.stack_top = state(prev)->mem.stack_top;
    incoming_mem.shadow_sp = state(prev)->mem.shadow_sp;
  }
  assert(all_code[PC.addr]);
  // abstract execution of current insn
  abstract_interp_insn (&incoming_mem, 
			&all_code[PC.addr]->decode);

  // merge with existing state at this location, if any
  if (state(PC)->visit_cnt == 1) {  // value is effectively top
    if (debug_data > 1) {
      cout << "  haven't been here before -- not merging state" << endl;
    }
    state(PC)->mem = incoming_mem;
  } else {
    if (debug_data > 1) {
      printf ("  have been here before (cnt == %d) -- merging state\n",
	      state(PC)->visit_cnt);
    }
    // can't merge these
    assert (incoming_mem.stack_top == state(PC)->mem.stack_top);
    assert (incoming_mem.shadow_sp == state(PC)->mem.shadow_sp);
    merge_state (&incoming_mem, &state(PC)->mem);
  }
  
  // ToDo: Once we deal with calls or if we decide to do some form of
  // path splitting, we would need to patch up outgoing contexts here.

  // possibly revise estimation of whether out-edges from here are
  // dead or not
  mark_dead_branches (state(PC), 
		      &all_code[PC.addr]->decode,
		      &all_code[PC.addr]->control);
  
  // did something happen?  changed is set -> fixpoint not yet reached
  bool changed;
  if (state(PC)->visit_cnt == 1) {
    changed = true;
  } else {
    changed = state_before_effect != *state (PC);
  }

  // moved clutter to separate routine
  if (debug_data > 1) print_stuff (PC, changed, &state_before_effect);

  return changed;
}

/*
 * insn at addr is the target of a new function call (to this context)
 */
static struct call_addr_ctx *add_new_call (abs_loc_t cur)
{
  if (debug_data > 1) cout << "  add_new_call to " << cur << endl;

  assert (cur.ctx < (code_ctxt_t)all_code[cur.addr]->state_cnt);
  assert (((state(cur)->calls != 0) && 
	   (state(cur)->num_calls > 0)) ||
	  ((state(cur)->calls == 0) && 
	   (state(cur)->num_calls == 0)));

  state(cur)->is_target = true;
  state(cur)->num_calls++;
  state(cur)->calls = (struct call_addr_ctx *) 
    xrealloc (state(cur)->calls, 
	      (size_t) state(cur)->num_calls * 
	      sizeof (struct call_addr_ctx));

  return &state(cur)->
    calls[state(cur)->num_calls-1];
}

/*
 * insn at addr has a new return associated with it (for this context)
 */
static void add_new_return (abs_loc_t cur, // function head
			    abs_loc_t ret) // return insn
{
  if (debug_data > 1) {
    cout << "  add_new_ret(" << cur << "," << ret << ")" << endl;
  }

  assert ((state(cur)->rets != NULL && 
	   state(cur)->num_rets > 0) ||
	  (state(cur)->rets == NULL && 
	   state(cur)->num_rets == 0));  
  assert (state(cur)->is_target);

  // make sure this return doesn't already exist
  for (int i=0; i<state(cur)->num_rets; i++) {
    if (state(cur)->rets[i].ret == ret) {
      if (debug_data > 1) {
        cout << "    already added return " << cur << " -> " << ret << endl; 
      }
      return;
    }
  }

  if (debug_data > 1) {
    cout << "    adding return " 
         << cur << " -> " << ret << " for the first time" << endl;
  }
  state(cur)->num_rets++;
  state(cur)->rets = (struct ret_addr_ctx *) 
    xrealloc (state(cur)->rets, 
	      (size_t) state(cur)->num_rets * 
	      sizeof (struct ret_addr_ctx));

  struct ret_addr_ctx *p = &state(cur)->rets[state(cur)->num_rets-1];
  p->ret = ret;
}

void add_call_info (abs_loc_t target,
		    abs_loc_t cur,
		    abs_loc_t succ,
		    abs_loc_t ret,
		    abs_loc_t head,
		    branch_idx_t idx)
{
  struct call_addr_ctx *call;

  lazy_create_contexts (target);
  
  // add the calling address/context to a list at the called
  // address/context
  call = add_new_call (target);
  
  if (debug_data > 1) {
    cout << "  add_call_info:" << endl
         << "   target " << target
         << " cur "    << cur 
         << " succ "   << succ
         << " ret "    << ret
         << " head "   << head
         << " idx "    << idx
         << endl;
  }
  
  call->call = cur;
  call->ret  = succ;
  call->stashed_ret  = ret;
  call->stashed_head = head;

  // remember the target context
  assert (target.ctx != NO_CTX);
  assert (all_code[cur.addr]->states != NULL);
  assert (state(cur)->out_edges.len >= (unsigned int) idx);

  state(cur)->out_edges.list[idx].target.ctx = target.ctx;
}

static int count_recursion (abs_loc_t calling,
			    code_addr_t called_addr, 
			    abs_loc_t *prev)
{
  int cnt = 0;
  abs_loc_t cur = calling;

  if (debug_data > 1) {
    printf ("  trying to find previous recursive calls to %x %s\n",
	    called_addr, addr_to_head_label (called_addr));
    printf ("  backwards call chain:\n");
  }

  while (state(cur)->head.addr != NO_CODE_ADDR) {

    abs_loc_t head = state(cur)->head;

    if (debug_data > 1) {
      printf ("    %x %d %s\n",
	      head.addr, head.ctx, addr_to_head_label (head.addr));
    }

    // did we find a call to the current function?
    if (called_addr == head.addr) {
      // save the address/context of the first recursive call we see
      if (cnt == 0) {
	*prev = head;
      }
      cnt++;
      if (!quiet) printf ("      found recursive call from %x %d %s -> %x %d %s\n",
			  calling.addr, calling.ctx, addr_to_head_label (calling.addr),
			  head.addr, head.ctx, addr_to_head_label (head.addr));
    }

    cur = state(head)->calls[0].call;
  }

  return cnt;
}

static void check_SP (abs_loc PC) 
{
  int shadow_sp = state(PC)->mem.shadow_sp;
  int real_sp = SP_val (&PC);
  if (sp_consistent[PC.addr] && shadow_sp != real_sp) {
    printf ("probable missing stack macro-insn: SP = %x but shadow SP = %x at %x.%d\n", 
	    real_sp, shadow_sp, PC.addr, PC.ctx);
    die ();
  }

  /*
   * indeterminate SP values not an inherent problem but for
   * now we want to see where they come from as soon as possible
   */
  if (ioport (PC, SPL_ADDR)->known != 0xff ||
      ioport (PC, SPH_ADDR)->known != 0xff) {
    cout << "oops: indeterminate SP value found at " 
         << PC
         << endl;
    die ();
  }
}

static void print_dataflow_step_stuff (abs_loc_t PC,
				       abs_loc_t prev,
				       abs_loc_t ret,
				       abs_loc_t head)
{
#if 0
  const char *s = addr_to_head_label (PC.addr);
  printf ("s = %p\n", s);
  fflush (stdout);
#endif
  
  assert( PC.ctx != NO_CTX );
  assert( state(PC) );
  cout << "dataflow step at " 
       << PC << "(" << addr_to_head_label (PC.addr) << ")"
       << " (prev " << prev << " " << addr_to_head_label (prev.addr) << ")"
       << " (head " << head << ")"
       << " (ret " << ret << ")"
       << endl
       << "  " << all_code[PC.addr]->decode.opcode 
       << endl
       << dec
       << "  branches in insn control info: "
       << "taken= "       << all_code[PC.addr]->control.index_taken
       << ", not_taken= " << all_code[PC.addr]->control.index_not_taken
       << ", len= "       << all_code[PC.addr]->control.edges_out.len
       << endl
       << "  cnt = " << state(PC)->visit_cnt << ";"
       << " branches out len= " << state(PC)->out_edges.len
       << endl
       << hex;
}

static void handle_return (abs_loc_t PC,
			   abs_loc_t head)
{
  assert (head.addr != NO_CODE_ADDR);
  state(head)->may_return = true;
  if (debug_data > 1) {
    cout << "  func with head " << head << " may return" << endl;
  }
  add_new_return (head, PC);

  // enqueue a work item from this return insn to every return
  // address/context (only one for context sensitive analysis,
  // right?)
  
  if (debug_data > 1) cout << "  head = " << head << endl;
  
  state_inst *s = state(head);
  
  if (debug_data > 1) printf ("  enqueuing all %d known return edges\n",
			      s->num_calls);
  
  for (int i=0; i < s->num_calls; i++) {
    
    if (debug_data > 1) printf ("    known_ret_edge from stash %x %d -> %x %d  (ret %x %d) (head %x %d)\n",
				PC.addr, PC.ctx,
				s->calls[i].call.addr, s->calls[i].call.ctx,
				s->calls[i].stashed_ret.addr, s->calls[i].stashed_ret.ctx,
				s->calls[i].stashed_head.addr, s->calls[i].stashed_head.ctx);
    
    enqueue_data (PC,
		  s->calls[i].ret,
		  s->calls[i].stashed_ret,
		  s->calls[i].stashed_head);
    
    
#if WCET_ANALYSIS
    // make sure this return doesn't already exist
    abs_loc ret_addr = s->calls[i].ret;
    int j;
    for (j=0; 
	 (j < state(ret_addr)->num_ret_targets)
	   && (state(ret_addr)->ret_targets[j] != PC);
	 j++) {
    }
    if (j >= state(ret_addr)->num_ret_targets) { // not found
      if (debug_data > 1) {
	cout << "    adding return " 
	     <<  ret_addr << " <- " << PC 
	     << " for the first time" << endl;
      }
      int n = state(ret_addr)->num_ret_targets + 1;
      state(ret_addr)->num_ret_targets = n;
      state(ret_addr)->ret_targets = (abs_loc*) 
	xrealloc (state(ret_addr)->ret_targets, 
		  (size_t) n * sizeof (abs_loc));
      state(ret_addr)->ret_targets[n-1] = PC;
    }
#endif
  }
}

abs_loc_t handle_call_first_time (abs_loc_t PC,
				  abs_loc_t ret,
				  abs_loc_t head,
				  branch_idx_t idx)
{
  abs_loc_t succ = nextPC (PC);
  code_ctxt_t t_addr = state(PC)->out_edges.list[idx].target.addr;
  abs_loc_t target;

  // check for recursion
  abs_loc_t rec;	    
  int recurse_cnt = count_recursion (PC, t_addr, &rec);
  assert (recurse_cnt == 0 || t_addr == rec.addr);
  
  if (recurse_cnt >= MAX_RECURSE_DEPTH) {
    // return top by killing this branch of the analysis
    if (debug_data > 2) {
      printf ("max recurse depth reached, killing this branch\n");
    }
    state(PC)->terminate_analysis = true;
    return NO_LOC;
  } else {
    target = abs_loc(t_addr, create_new_context (t_addr));
  }

  state(PC)->out_edges.list[idx].target.ctx = target.ctx;
  
  if (print_callgraph) {
    fprintf (callgraph_file, "  \"%s\" -> \"%s\";\n",
	     addr_to_head_label (PC.addr),
	     addr_to_head_label (target.addr));
    fprintf (df_cg_file, "  \"%s__%d\" -> \"%s__%d\";\n",
	     addr_to_head_label (PC.addr), PC.ctx,
	     addr_to_head_label (target.addr), target.ctx);
  }
  
  if (debug_data > 1) {
    cout << "  first time seeing call "
	 << PC << " (" << addr_to_head_label (PC.addr) << ") -> "
	 << target << " (" << addr_to_head_label (target.addr) << ") "
	 << "(ret "  << ret << ") "
	 << "(head " << head << ")"
	 << endl;
  }

  add_call_info (target,
		 PC,
		 succ,
		 ret,
		 head,
		 idx);
   
  // enqueue a work item from each currently-known return in
  // the function to the call's successor
  int i;
  state_inst *s = state(target);
  
  if (debug_data > 1) {
    printf ("  seeing call for the first time, so enqueuing back-edges from all %d known rets\n",
	    s->num_rets);
  }
	  
  for (i=0; i < s->num_rets; i++) {
    
    if (debug_data > 1) {
      cout << "    back-edge from " << s->rets[i].ret 
	   << " -> "                << succ 
	   << endl;
    }
	    
    enqueue_data (s->rets[i].ret,
		  succ,
		  ret,
		  head);
  }

  // handle "higher order functions" like TOS_post

  struct ho_fn_t *hofn = lookup_ho_fn (target.addr);    
  if (hofn) {

    if (debug_data > 0) {
      printf ("adding a deferred call to add a tinyos task to %x (%s)\n",
	      target.addr, lookup_text_symbol (target.addr));
    }

    // get the main loop to later call add_tinyos_task_data
    call_node_t call;
    call.calling_pc   = PC;
    call.called_addr = target.addr;
    call_queue.push (call);
  }


  return target;
}

abs_loc_t handle_call_subsequent (abs_loc_t PC,
				  branch_idx_t idx)
{
  // seen it before -- use previously allocated context	  
  abs_loc_t target = state(PC)->out_edges.list[idx].target;
  assert (target.ctx != NO_CTX);
  
  if (!state(target)->is_target) {
    cout << "Target: " << target << " / "
	 << "idx: " << idx << " / "
	 << "len: " << state(PC)->out_edges.len << " / "
	 << endl;
  }
  assert (state(target)->is_target);

  return target;
}


void drain_cf (void)
{
  while (true) {
    bool cont = control_flow_step ();
    if (!cont) break;
  }
}

void handle_indirect_call (abs_loc_t PC,
			   abs_loc_t ret,
			   abs_loc_t head)
{
  if (is_scheduler (head.addr)) {
    
    // case 1 -- this icall is in the tinyos scheduler, so we'll
    //           find tasks later as we hit calls to tos_post
    
    if (state(PC)->visit_cnt == 1) {
      if (debug_data > 0)
	printf ("seem to have found the TinyOS scheduler at %x\n", PC.addr);
      // HACK: scheduler not permitted to run in contexts other than zero!!
      assert (PC.ctx == 0);
      add_scheduler (PC.addr, ret, head);
    } else {
      if (find_scheduler (PC.addr) == 0) {
	  printf("Yoiks, we somehow managed to ignore icall at 0x%x the first time we visited it.\n", 
		 PC.addr);
	  die();
      }	
    }
    
  } else {
    
    // case 2 -- this icall is not in the scheduler, in which case
    //           we require a concrete address to call
    
    if (debug_data > 0) {
      printf ("found indirect call from %x %s\n", 
	      head.addr, 
	      lookup_text_symbol (head.addr));
    }
      
    mem_loc_t lo = state(PC)->mem.core[reg_to_mem(30)];
    mem_loc_t hi = state(PC)->mem.core[reg_to_mem(31)];
   
    if (lo.known != 0xff || hi.known != 0xff) {
      
      printf ("**************************************************************\n");
      printf ("OOPS -- found icall at %x with inconclusive arguments\n",
	      PC.addr);
      printf ("(%x.%x %x.%x)!\n",
	      lo.known, lo.val,
	      hi.known, hi.val);
      printf ("can't continue analyzing\n");
      printf ("**************************************************************\n");
      die();
    }
    
    code_addr_t icall_target = pair_to_addr (lo.val, hi.val);
    
    // is this a new icall from this address?
    // FIXME -- should do this using value set analysis!
    
    bool found = false;
    for (branch_idx_t idx=0; idx< (branch_idx_t)state(PC)->out_edges.len; idx++) {
      if (state(PC)->out_edges.list[idx].target.addr == icall_target) {
	found = true;
      }
    }
    
    if (!found) {

      if (debug_data > 0) {
	printf ("new indirect call target is %x (%s)\n",
		icall_target,
		lookup_text_symbol (icall_target));
      }
	
      add_branch (PC.addr, icall_target, false, (cost_t)4);
      
      // FIXME hacky experiment -- do a little control flow here in
      // the middle of the dataflow
      drain_cf ();      
    }      
  }
}

void handle_calls (abs_loc_t PC,
		   abs_loc_t ret,
		   abs_loc_t head)
{
  abs_loc_t succ = nextPC (PC);

  // indirect call stuff
      
  if ((all_code[PC.addr]->decode.opcode == AVR_icall ||
       all_code[PC.addr]->decode.opcode == AVR_eicall)) {
    handle_indirect_call (PC, ret, head);
  }

  /*
   * look at each target for this call -- only indirect calls will
   * go multiple places
   */
  for (branch_idx_t idx=0; idx< (branch_idx_t)state(PC)->out_edges.len; idx++) {

    abs_loc_t target; 
       
    if (state(PC)->visit_cnt == 1) {	    
      target = handle_call_first_time (PC, ret, head, idx);
    } else {
      target = handle_call_subsequent (PC, idx);
    } 

    // NO_LOC only happens when we decide a recursion has reached its limit
    if (target != NO_LOC) {
      enqueue_data (PC, 
		    target, 
		    succ,
		    target);
    }
  }
}

bool dataflow_step (void)
{
  if (debug_data > 1) {
    cout << endl << endl;
  }

  abs_loc_t PC;
  abs_loc_t prev;
  abs_loc_t ret;
  abs_loc_t head;

  dequeue_data (&prev, &PC, &ret, &head);

  if (PC.addr == NO_CODE_ADDR) { return false; }

  if (debug_data > 1) {
    print_dataflow_step_stuff (PC, prev, ret, head);
  }

  assert (all_code[PC.addr] != NULL);
  assert (all_code[PC.addr]->touched);
  assert (PC.ctx != NO_CTX);
  assert (PC.ctx < (code_ctxt_t)all_code[PC.addr]->state_cnt);
  assert (head.ctx != NO_CTX);
  assert (head.ctx < (code_ctxt_t)all_code[PC.addr]->state_cnt);
  assert (state(PC)->visit_cnt >= 0);

  state(PC)->visit_cnt++;

  if (state(PC)->visit_cnt == 1) {
    if (!(state(PC)->head.addr == NO_CODE_ADDR)) {
      cout << "PC=" << PC << "=" << state(PC)->head << endl;
    }
    assert (state(PC)->head.addr == NO_CODE_ADDR);
    state(PC)->head = head;
  } else {
    if (state(PC)->head != head) {
      // FIXME -- Why?
#if VERBOSE_CTRL_FLOW
      if (strncmp ("__div", all_code[head.addr]->decode.label, 5) != 0) {
	cout << "warning: address " 
	     << PC << " (" << all_code[head.addr]->decode.label << ")"
	     << " belongs to more than one function ("
	     <<    state(PC)->head << " and " << head << ")" << endl;
      }
#endif
    }
  }

  if (state(PC)->terminate_analysis) {
    if (debug_data > 1) {
      printf ("oops, no need to analyze %x\n", PC.addr);
    }
    return true;
  }

  // most real work happens here
  bool changed = dataflow_effect (PC, prev);

  check_SP (PC);

  if (all_code[PC.addr]->decode.opcode == AVR_ret) {
    handle_return (PC, head);
  }
  
  if (!changed) {
    if (debug_data > 1) {
      cout << "  NO CHANGE -- bailing without creating more work items" << endl;
    }
    return true;
  }

  if (debug_data > 1) {
    printf ("  following %d out c.f. edge(s)\n",
	    state(PC)->out_edges.len);
    print_verbose_dataflow_ctx (PC, stdout, cout);
  }  

  switch (all_code[PC.addr]->decode.opcode) {
    
  case AVR_call:
  case AVR_rcall:
    assert (state(PC)->out_edges.len <= 1);
    /*lint -fallthrough */
  case AVR_icall:
  case AVR_eicall:
    handle_calls (PC, ret, head);
    break;
    
  case AVR_ret:
    // already handled 
    break;
    
  case AVR_eijmp:
    printf ("oops -- eijmp can't be analyzed yet\n");
    die();
    break;

  case AVR_ijmp:
    {
      state_inst *s = state(PC);

      mem_loc_t r30 = s->mem.core[reg_to_mem(30)];
      mem_loc_t r31 = s->mem.core[reg_to_mem(31)];

      if (r30.known != 0xff || r31.known != 0xff) {
        printf ("oops -- found ijmp at %x.%d with inconclusive arguments!\n",
		PC.addr, PC.ctx);
        die ();
      } 
      
      code_addr_t target_addr = pair_to_addr (r30.val, r31.val);

      if (!quiet) printf ("cool: analyzed ijmp to %x\n", target_addr);

      // FIXME -- the below says that an icall can only have a single
      // target; of course we can do better here if we have a
      // value-set analysis

      if (state(PC)->out_edges.len == 0) {
	(void) add_branch (PC.addr, target_addr, false, (cost_t)2);      
	drain_cf ();
      } else {
	assert (state(PC)->out_edges.len == 1);
      }
    }
    // NOTE!! deliberate fallthrough!!
    
  default:
    // not a call or a return -- just follow all out edges
    {
      unsigned int i;
      
      for (i=0; i < state(PC)->out_edges.len; i++) {
        abs_loc_t n = abs_loc(state(PC)->out_edges.list[i].target.addr,PC.ctx);

	if (debug_data > 1) {
	  printf ("    edge %d to %x %s\n",
		  i, 
		  state(PC)->out_edges.list[i].target.addr,
		  (state(PC)->out_edges.list[i].dead) ? 
		  "is dead" : "isn't dead");
	}
	
	// follow only live branches -- context stays the same
	if (!state(PC)->out_edges.list[i].dead) {
	  enqueue_data (PC, 
			n, 
			ret,
			head);
	}
      }
    }
  }
  
  return true;
}

void dataflow_setup (void)
{
}  

/*
 * find branches that cannot be taken in any context
 */
void report_dead_branches (void)
{
  code_addr_t addr;
  code_ctxt_t ctx;
  unsigned int branch;
  unsigned int len;
  bool dead;

  if (!quiet) printf ("\nbranches that are dead in all contexts:\n");

  for (addr=0; addr<(code_addr_t)code_size; addr++) {

    if (all_code[addr] == NULL) continue;
    if (all_code[addr]->states == NULL) continue;

    len = all_code[addr]->control.edges_out.len;

    for (branch=0; branch<len; branch++) {
      dead = true;

      if (all_code[addr]->control.edges_out.list[branch].target.addr == 0) {
	printf ("woah -- suspicious branch from %x to 0\n",
		addr);
      }

      for (ctx=(code_ctxt_t)0; ctx< (code_ctxt_t)all_code[addr]->state_cnt; ctx++) {

	assert (len == all_code[addr]->states[ctx].out_edges.len);

	if (all_code[addr]->control.edges_out.list[branch].target.addr !=
	    all_code[addr]->states[ctx].out_edges.list[branch].target.addr) {
	  printf ("woah -- at %x target %x is not the same as target %x in context %d\n",
		  addr,
		  all_code[addr]->control.edges_out.list[branch].target.addr,
		  all_code[addr]->states[ctx].out_edges.list[branch].target.addr,
		  ctx);
	  die ();
	}

	// it's not dead if anyone says it's not dead
	if (!all_code[addr]->states[ctx].out_edges.list[branch].dead) {
	  dead = false;
	}

      }

      if (dead) {
	if (!quiet) printf ("  0x%04x -> 0x%04x\n",
			    addr, all_code[addr]->control.edges_out.list[branch].target.addr);
	all_code[addr]->control.edges_out.list[branch].dead = true;
      } else {
	all_code[addr]->control.edges_out.list[branch].dead = false;
      }

    }
  }
}

static void print_mem_loc (struct mem_loc_t m)
{
  int i;
  for (i=7; i>=0; i--) {
    if (is_bit_set (m.known, i)) {
      if (is_bit_set (m.val, i)) {
	printf ("1");
      } else {
	printf ("0");
      }
    } else {
      printf ("?");
    }
  }
}

static void print_reg_pair (memory *m, int first)
{
  printf ("[");
  print_mem_loc (m->core[reg_to_mem(first+1)]);
  printf (" ");
  print_mem_loc (m->core[reg_to_mem(first)]);
  printf ("]  ");
}

static void print_x (memory *m)
{
  print_reg_pair (m, 26);
}

static void print_y (memory *m)
{
  print_reg_pair (m, 28);
}

static void print_z (memory *m)
{
  print_reg_pair (m, 30);
}

void report_stores (void)
{
  code_addr_t addr;

  if (!quiet) printf ("\nstores:\n");

  for (addr=0; addr<(code_addr_t)code_size; addr++) {

    if (all_code[addr] == NULL) continue;
    if (all_code[addr]->states == NULL) continue;

    switch (all_code[addr]->decode.opcode) {
    case AVR_sts:
      assert (all_code[addr]->decode.k_set);
      {
	data_addr_t target = all_code[addr]->decode.k;
	/*
	 * direct stores are only permitted into these segments: data,
	 * bss, extended i/o
	 */
	if ((target >= platform->sram.offset &&
	     target < platform->sram.offset + (data_addr_t)(data_seg_size + bss_seg_size)) ||
	    (target >= platform->extended_io.offset &&
	     target < (platform->extended_io.offset + (data_addr_t) platform->extended_io.len))) {
	  if (debug_data > 1) {
	    printf ("  %s to %x found at location %x -- probably safe\n", 
		    opcode_to_string (all_code[addr]->decode.opcode),
		    target,
		    addr);
	  }
	} else {
	  printf ("  %s to %x found at location %x -- probably UNsafe\n", 
		  opcode_to_string (all_code[addr]->decode.opcode),
		  target,
		  addr);
	  // assert (false);
	}
      }
      break;
    case AVR_st1x:
    case AVR_st2x:
    case AVR_st3x:
    case AVR_st1y:
    case AVR_st2y:
    case AVR_st3y:
    case AVR_std1y:
    case AVR_st1z:
    case AVR_st2z:
    case AVR_st3z:
    case AVR_std1z:
      if (!quiet) {
	printf ("  ");
	
	switch (all_code[addr]->decode.opcode) {
	case AVR_st1x:
	  printf ("st X");
	  break;
	case AVR_st2x:
	  printf ("st X+");
	  break;
	case AVR_st3x:
	  printf ("st -X");
	  break;
	case AVR_st1y:
	  printf ("st Y");
	  break;
	case AVR_st2y:
	  printf ("st Y+");
	  break;
	case AVR_st3y:
	  printf ("st -Y");
	  break;
	case AVR_std1y:
	  printf ("std Y+d");
	  break;
	case AVR_st1z:
	  printf ("st Z");
	  break;
	case AVR_st2z:
	  printf ("st Z+");
	  break;
	case AVR_st3z:
	  printf ("st -Z");
	  break;
	case AVR_std1z:
	  printf ("std Z+d");
	  break;
	default:
	  break;
	}
	
	printf (" found at location %x -- potentially unsafe\n", addr);
	code_ctxt_t j;
	
	printf ("    addrs: ");
	
	for (j=(code_ctxt_t)0; j < (code_ctxt_t)all_code[addr]->state_cnt; j++) {
	  
	  state_inst tmp_state = all_code[addr]->states[j];
	  
	  if (tmp_state.visit_cnt == 0) continue;
	  
	  switch (all_code[addr]->decode.opcode) {
	  case AVR_st1x:
	  case AVR_st2x:
	  case AVR_st3x:
	    print_x (&all_code[addr]->states[j].mem);
	    break;
	  case AVR_st1y:
	  case AVR_st2y:
	  case AVR_st3y:
	  case AVR_std1y:
	    print_y (&all_code[addr]->states[j].mem);
	    break;
	  case AVR_st1z:
	  case AVR_st2z:
	  case AVR_st3z:
	  case AVR_std1z:
	    print_z (&all_code[addr]->states[j].mem);
	    break;
	  default:
	    break;
	  }
	  
	}
	
	printf ("\n");
      }
      break;
    default:
      ;
    }
  }
}

// Convert old-style branch_list to new-style set
set<abs_loc_t,MAX_PCS> edges_to_set(const branch_list &bs) {
  set<abs_loc_t,MAX_PCS> s;
  for( unsigned i=0; i<bs.len; ++i) {
    if (!bs.list[i].dead) {
      s.insert(bs.list[i].target);
    }
  }
  return s;
}

void check_dataflow_invariants (void)
{
  code_addr_t addr;
  code_ctxt_t ctx;
  int x;

  for (addr=0; addr < (code_addr_t)code_size; addr++) {

    if (all_code[addr] == NULL) continue;
    if (all_code[addr]->states == NULL) continue;
    
    if (debug_data > 2) printf ("  checking invariants at %x\n", addr);
    
    for (ctx = (code_ctxt_t)0; ctx < (code_ctxt_t)all_code[addr]->state_cnt; ctx++) {

      assert (is_canonical_state (&all_code[addr]->states[ctx]));

      if (all_code[addr]->states[ctx].num_calls != 0) {
	assert (all_code[addr]->states[ctx].calls != NULL);
	assert (all_code[addr]->states[ctx].is_target);
	for (x=0; x<all_code[addr]->states[ctx].num_calls; x++) {
	  code_addr_t call_addr = all_code[addr]->states[ctx].calls[x].call.addr;
	  assert (all_code[call_addr]->decode.opcode == AVR_call ||
		  all_code[call_addr]->decode.opcode == AVR_rcall ||
		  all_code[call_addr]->decode.opcode == AVR_icall);
	}
      } else {
	assert (all_code[addr]->states[ctx].calls == NULL);
	assert (!all_code[addr]->states[ctx].is_target);
      }

      if (all_code[addr]->states[ctx].num_rets != 0) {
	assert (all_code[addr]->states[ctx].is_target);
	assert (all_code[addr]->states[ctx].rets != NULL);

	for (x=0; x<all_code[addr]->states[ctx].num_rets; x++) {
	  code_addr_t ret_addr = all_code[addr]->states[ctx].rets[x].ret.addr;
#if 0
          cout << "Checking return from function " << abs_loc(addr,ctx) 
               << " " << all_code[addr]->states[ctx].rets[x].ret
               << ": " << all_code[ret_addr]->decode.opcode 
               << endl;
#endif
	  assert (all_code[ret_addr]->decode.opcode == AVR_ret);
	}

      } else {
	//assert (!all_code[addr]->states[ctx].is_target);
	//assert (!all_code[addr]->states[ctx].rets);
      }

    }

  }
}
