#include "stacktool.h"

// global flag for printing global IPG in graphviz format
bool print_ipg = false;

// indexes into stack queues
typedef unsigned int slot_t;
const slot_t NO_SLOT = ((slot_t)-99);
const slot_t DEFAULT_DEPTH_QUEUE_SIZE = 100000;

// Used when looking for worst address so far
// (if it has this value, it's not initialized yet)
const code_addr_t NO_CODE_ADDR2 = ((code_addr_t)-99);

enum stack_cmd {
  // Within a flow, we walk forwards through the control flow and
  // then retrace our steps backwards to clean things up again
  ST_FORWARD = 2925,
  ST_BACKWARD,

  // A "flow" is either main or an irq handler
  ST_ADD_FLOW,  // start handler
  ST_DEL_FLOW,  // stop handler (decrement reentrancy count)
  ST_NO_CMD
};

static const char * cmd_to_str (enum stack_cmd cmd)
{
  switch (cmd) {
  case ST_FORWARD:
    return "FORWARD";
  case ST_BACKWARD:
    return "BACKWARD";
  case ST_ADD_FLOW:
    return "ADD_FLOW";
  case ST_DEL_FLOW:
    return "DEL_FLOW";
  case ST_NO_CMD:
    return "NO_CMD";
  default:
    die();
  }
}

struct instance_cnt_t {
  int cnt[MAX_VECTORS];
};

instance_cnt_t uninit_instance_cnt, zero_instance_cnt;

struct stack_queue_elt {
  enum stack_cmd cmd;
  abs_loc_t cur;
  abs_loc_t prev;
  int depth;
  int path_len;
  slot_t prev_slot;  // for back-tracing
  instance_cnt_t instance_cnt;
};

struct stack_queue_t {
  struct stack_queue_elt q[DEFAULT_DEPTH_QUEUE_SIZE];
  slot_t index;
};

// worklist
// Note: unlike many worklists, the order of elements matters.
// For example, we will push a DEL_FLOW below an ADD_FLOW so that
// the flow gets deleted as soon as we are done with this flow.
static struct stack_queue_t stack_queue;

static slot_t enqueue_stack (const struct stack_queue_elt *q)
{
  if (debug_depth > 1) {
    cout << "    enqueue_stack"
         << ": cmd= " << cmd_to_str (q->cmd)
         << ", PC= " << q->cur
         << ", depth= " << dec << q->depth << hex
         << ", ints= " << (safe_ints_enabled(q->cur) ? "+" : "-")
         << endl;
  }

  stack_queue.q[stack_queue.index] = *q;

  slot_t idx = stack_queue.index;
  stack_queue.index++;
  assert (stack_queue.index < DEFAULT_DEPTH_QUEUE_SIZE);

  return idx;
}

static slot_t dequeue_stack (struct stack_queue_elt *dq)
{
  if (stack_queue.index == (slot_t)0) {
    dq->cmd = ST_NO_CMD;
    return (slot_t)-1;
  }
  stack_queue.index--;
  *dq = stack_queue.q[stack_queue.index];
  return stack_queue.index;
}

/*
 * return true if interrupts are provably enabled at this program
 * point
 *
 * Only used in generating graphviz callgraph, not for 'internal' purposes.
 */
bool ints_enabled (abs_loc_t cur)
{
  return 
    is_bit_set (ioport(cur,SREG_ADDR)->known, I_BIT) &&
    is_bit_set(ioport(cur,SREG_ADDR)->val, I_BIT);
}

/*
 * return true unless it's impossible that interrupts are enabled at
 * this point in the program
 */
bool safe_ints_enabled (abs_loc_t cur)
{
  if (cur.addr == NO_CODE_ADDR) return false;

  return 
    !is_bit_set (ioport(cur,SREG_ADDR)->known, I_BIT) ||
    is_bit_set (ioport(cur,SREG_ADDR)->val, I_BIT);
}

bool int_known (abs_loc_t cur, int vec)
{
  if (cur.addr == NO_CODE_ADDR) return false;

  data_addr_t a = platform->int_vecs[vec].enable.addr;
  bit_t b = platform->int_vecs[vec].enable.bit;
  data_kno_t k = memry (cur,a)->known;
  return is_bit_set (k, b);
}

bool int_enabled (abs_loc_t cur, int vec)
{
  if (cur.addr == NO_CODE_ADDR) return false;

  return 
    is_bit_set (memry(cur,platform->int_vecs[vec].enable.addr)->val,
		platform->int_vecs[vec].enable.bit);
}

static bool safe_int_enabled (abs_loc_t cur, int vec) {
  return (!int_known (cur, vec) 
          || int_enabled (cur, vec)
#if 1
          // Use this if you're worried that we're missing interrupts
          // because they're being enabled inside an interrupt handler
          // (It doesn't seem to affect results).
          || indirectly_enabled[vec]
#endif
          );
}

static void mark_path (int depth)
{
  slot_t slot = stack_queue.index;

  while (slot != NO_SLOT) {
    code_addr_t addr = stack_queue.q[slot].cur.addr;
    assert (slot != NO_SLOT && slot <= stack_queue.index);
    all_code[addr]->wc_path_depth = 
      max<int> (all_code[addr]->wc_path_depth, depth);
    slot = stack_queue.q[slot].prev_slot;
  }
}

// decluttering do_stack
static void print_stuff (struct stack_queue_elt *dq)
{
  if (all_code[dq->cur.addr]->decode.label != NULL) {
    cout << all_code[dq->cur.addr]->decode.label << ":" << endl;
  }
  if (debug_depth==1) {
    cout << "ints=" << (safe_ints_enabled (dq->cur) ? "+" : "-")
	 << ", old_wc=" << dec << state(dq->cur)->wc_stack_depth << hex
	 << ", depth=" << dec << dq->depth << hex
	 << ", paths=" << dec << state(dq->cur)->path_cnt << hex
	 << "\t";
  }    
  cout << dq->cur;
  cout << ": ";
  avr_disasm_insn(dq->cur.addr, stdout);
  cout << endl;
  if (debug_depth>1) {
    cout << "\tints=" << (safe_ints_enabled (dq->cur) ? "+" : "-")
	 << ", old_worst_depth= " << dec << state(dq->cur)->wc_stack_depth << hex
	 << ", depth= " << dec << dq->depth << hex
	 << ", path_cnt= " << dec << state(dq->cur)->path_cnt << hex
	 << endl;
  }
}

static int get_sp (abs_loc_t loc)
{
  assert (ioport (loc, SPH_ADDR)->known == 0xff &&
	  ioport (loc, SPL_ADDR)->known == 0xff);

  return 
    (ioport (loc, SPH_ADDR)->val << 8) + 
    ioport (loc, SPL_ADDR)->val;
}     

static int do_stack (struct stack_queue_elt *dq)
{
  bool live_out, may_return;
  int delta;
  slot_t slot;
  
  assert (dq->depth >= 0);  
  assert (state(dq->cur)->path_cnt >= 0);

  int cur_size = state(dq->cur)->mem.stack_top - get_sp (dq->cur);
  int prev_size;
  if (dq->prev.addr == NO_CODE_ADDR) {
    prev_size = 0;
  } else {
    prev_size = state(dq->prev)->mem.stack_top - get_sp (dq->prev);
  }
  delta = cur_size - prev_size;    

  dq->depth += delta;

  if (debug_depth) {
    print_stuff (dq);
  }

  // FIXME -- ignore SP where known to be inconsistent

  if (dq->depth <= state(dq->cur)->wc_stack_depth) {
    if (debug_depth>1) printf ("(stop -- no inc)\n");
    return dq->depth;
  }

  if (state(dq->cur)->terminate_analysis) {
    if (debug_depth>1) printf ("(stop -- path_cnt=%d)\n", 
                               state(dq->cur)->path_cnt);
    return dq->depth;
  }

  if (debug_depth>1) printf ("(go)\n");

  /*
   * now we're committed, so update the depth and increment the path count
   */
  if (mark_deepest_path) {
    mark_path (dq->depth);
  }

  state(dq->cur)->wc_stack_depth = max<int> (state(dq->cur)->wc_stack_depth,
					     dq->depth);

  state(dq->cur)->path_cnt++;

  {
    struct stack_queue_elt q = {
      ST_BACKWARD, 
      dq->cur, 
      abs_loc(NO_CODE_ADDR,NO_CTX),
      0, 
      dq->path_len, 
      dq->prev_slot,
      uninit_instance_cnt,
    };
    slot = enqueue_stack (&q);
  }
    
  // sanity check -- returning interrupts should leave a clean stack
  if (all_code[dq->cur.addr]->decode.opcode == AVR_reti &&
      dq->depth != 0) {
    die ();
  }

  // XXX

  live_out = false;
  may_return = false;
  if (all_code[dq->cur.addr]->decode.opcode != AVR_ret) {
    for (unsigned int i=0; i<state(dq->cur)->out_edges.len; i++) {
      abs_loc_t n;

      n.addr = state(dq->cur)->out_edges.list[i].target.addr;
      if (avr_is_call(all_code[dq->cur.addr]->decode.opcode)) {
	n.ctx = state(dq->cur)->out_edges.list[i].target.ctx;
      } else {
	n.ctx = dq->cur.ctx;
      }

      assert (n.ctx != NO_CTX);
      
      // don't follow dead branches
      if (state(dq->cur)->out_edges.list[i].dead) {
	if (debug_depth>1) printf ("  dead branch to %x\n", n.addr);
      } else {
	if (debug_depth>1) printf ("  live branch to %x\n", n.addr);
	{
	  struct stack_queue_elt q = {
	    ST_FORWARD, 
	    n,
	    dq->cur,
	    dq->depth,
	    dq->path_len+1, 
	    slot,
	    dq->instance_cnt,
	  };
	  (void) enqueue_stack (&q);
	}				

	live_out = true;
	if (state(n)->may_return) {
	  may_return = true;
	}

      }
    }

  }

  /*
   * hack: procedure calls don't actually transfer control to their
   * successor insns, but here we act like they do --- but only for
   * cases where we can't prove that the call never returns
   */
  if (live_out && avr_is_call(all_code[dq->cur.addr]->decode.opcode)) {

    if (may_return) {     
      struct stack_queue_elt q = {
	ST_FORWARD, 
	nextPC(dq->cur), 
	dq->cur,
	dq->depth,
	dq->path_len+1, 
	slot,
	dq->instance_cnt,
      };
      (void) enqueue_stack (&q);
    } else {
#if 1 || VERBOSE_CTRL_FLOW
      if (!quiet) {
	printf ("  WOAH: function call at %x %s %d provably doesn't return!\n",
		dq->cur.addr, 
		addr_to_head_label (dq->cur.addr),
		dq->cur.ctx);
      }
#endif
    }
  }

  return dq->depth;
}


static int drain_stack_queue (int *max_depth)
{
  int steps = 0;

  while (true) {

    if (debug_depth>1) printf ("\n");

    steps++;
    struct stack_queue_elt dq;
    dequeue_stack (&dq);

    switch (dq.cmd) {

    case ST_DEL_FLOW: // kill off an irq handler
      {
	unsigned int vec = (unsigned int) dq.cur.addr / 4;
	// currently true but not necessarily in the long run
	assert (dq.cur.addr == platform->int_vecs[vec].start_addr);
	dq.instance_cnt.cnt[vec]--;
      }
      break;

    case ST_ADD_FLOW: // start up an irq handler
      {
	unsigned int vec = (unsigned int) dq.cur.addr / 4;
	// currently true but not necessarily in the long run
	assert (dq.cur.addr == platform->int_vecs[vec].start_addr);
	dq.instance_cnt.cnt[vec]++;
      }
      /*lint -fallthrough */
    case ST_FORWARD: {

      if (debug_depth>1) {
        cout << cmd_to_str (dq.cmd) << ": "
             << "addr= " << dq.cur << ", "
             << "depth= " << dq.depth << ", "
             << "max_depth= " << *max_depth 
             << endl;
      }

      assert (all_code[dq.cur.addr] != NULL);
      assert (dq.cur.ctx != NO_CTX);
      assert (all_code[dq.cur.addr]->states != NULL);

      /*
       * most work is done here 
       */
      int depth_here = do_stack (&dq);

      if (depth_here > *max_depth) {
	*max_depth = depth_here;
      }

      break;
    }
    case ST_BACKWARD:
      if (debug_depth>1) {
	cout << "ST_BACKWARD " << dq.cur;
      }
      state(dq.cur)->path_cnt--;
      break;

    case ST_NO_CMD:
      return steps;

    default:
      die ();

    }
  }
}

static int wc_depth[MAX_VECTORS+1];
static abs_loc_t wc_depth_loc[MAX_VECTORS+1];
static int wc_depth_to_int[MAX_VECTORS+1][MAX_VECTORS+1];
static abs_loc_t wc_depth_to_int_loc[MAX_VECTORS+1][MAX_VECTORS+1];

// propagate per-context information to the global results field
static void collect_results (int vec)
{
  for (code_addr_t addr=0; addr < (code_addr_t) code_size; addr++) {
    if (all_code[addr] == NULL) continue;
    for (code_ctxt_t ctx = (code_ctxt_t)0; 
	 ctx < (code_ctxt_t)all_code[addr]->state_cnt; 
	 ctx++) {
      abs_loc_t loc = abs_loc_t (addr, ctx);
      assert (state(loc)->path_cnt == 0);
      
      int d = state(loc)->wc_stack_depth;
      
      if (d > wc_depth[vec]) {
	wc_depth[vec] = d;
	wc_depth_loc[vec] = loc;
      }
      
      for (unsigned int vec2=1; vec2<platform->num_vectors; vec2++) {
	if (safe_ints_enabled (loc) && safe_int_enabled(loc, vec2)) {
	  
	  if (d > wc_depth_to_int[vec][vec2]) {
	    wc_depth_to_int[vec][vec2] = d;
	    wc_depth_to_int_loc[vec][vec2] = loc;
	  }
	  
	}
      }

      all_code[addr]->wc_stack_depth = 
	max<int> (state(loc)->wc_stack_depth, all_code[addr]->wc_stack_depth);

      state(loc)->wc_stack_depth = NEG_INFINITY;
    }
  }
}

static void setup_stack_scan (void)
{
  for (int vec=0; vec < MAX_VECTORS+1; vec++) {
    zero_instance_cnt.cnt[vec] = 0;
    wc_depth[vec] = NEG_INFINITY;
    wc_depth_loc[vec] = NO_LOC;
    for (int j=0; j<MAX_VECTORS+1; j++) {
      wc_depth_to_int[vec][j] = NEG_INFINITY;
      wc_depth_to_int_loc[vec][j] = NO_LOC;
    }
  }

  for (code_addr_t addr=0; addr < (code_addr_t) code_size; addr++) {
    if (all_code[addr] == NULL) continue;
    all_code[addr]->wc_path_depth = NEG_INFINITY;
    for (code_ctxt_t ctx = (code_ctxt_t)0; 
	 ctx < (code_ctxt_t)all_code[addr]->state_cnt; 
	 ctx++) {
      abs_loc_t loc = abs_loc_t (addr, ctx);
      state (loc)->wc_stack_depth = NEG_INFINITY;
    }
  }
}

struct stacked_ints {
  int depth[MAX_VECTORS+1];
};

static FILE *dotfile = NULL;
static int edges = 0;

static void add_edge (int vec, int len, int int_cnt,
		      char *in_str, char *new_str)
{
  assert (int_cnt < MAX_VECTORS);

  if (wc_depth[vec] <= PC_SIZE) return;

  edges++;

  fprintf (dotfile, "  \"%s\" [label=\"int%d (%d)\"]\n", 
	   new_str, vec, wc_depth[vec]);

  fprintf (dotfile, "  \"%s\" -> \"%s\" [label=\"%d\"];\n",
	   in_str, new_str, len);
}

static int global_depth_helper (int vec, 
				int in_depth,
				struct stacked_ints stack,
				int int_cnt, 
				char *in_str)
{
  stack.depth[vec]++;

  int out_depth = wc_depth[vec];

  for (unsigned int i=1; i<platform->num_vectors; i++) {

    if (wc_depth[i] == NEG_INFINITY) continue; 
    if (wc_depth_to_int[vec][i] == NEG_INFINITY) continue; 
    if (stack.depth[i] >= int_info[i].max_reenter) continue; 

#if 0
    // FIXME -- broken but useful for suppressing false positive edges
    if (vec > 0 && wc_depth_to_int[vec][i] <= 4) continue; 
#endif

    char *new_str = NULL;

    if (print_ipg) {
      new_str = (char *) xmalloc (255);
      sprintf (new_str, "%s_%d", in_str, i);
      add_edge (i, wc_depth_to_int[vec][i], int_cnt, in_str, new_str);
    }

    int new_depth = global_depth_helper (i, in_depth, stack, int_cnt+1, new_str);

    out_depth = max<int> (out_depth, wc_depth_to_int[vec][i] + new_depth);
			  
  }

  if (print_ipg) free (in_str);

  return in_depth + out_depth;
}

static int global_depth (void)
{
  struct stacked_ints stack;

  for (unsigned int i=0; i<platform->num_vectors; i++) {
    stack.depth[i] = 0;
  }

  char *str = NULL;

  if (print_ipg) {
    dotfile = fopen ("graph.dot", "w");
    assert (dotfile);
    fprintf (dotfile, "digraph ints {\n");
    fprintf (dotfile, "  origin [shape=plaintext,label=\"\"];\n");
    fprintf (dotfile, "  \"0\" [label=\"main (%d)\"]\n", wc_depth[0]);

    str = (char *) xmalloc (255);
    sprintf (str, "0");
  }
  
  int all_max = global_depth_helper (0, 0, stack, 0, str);

  if (print_ipg) {
    fprintf (dotfile, "}\n");
    fclose (dotfile);
    printf ("\n%d total edges in the IPG\n", edges);
  }

  return all_max;
}

static void check_safety (int depth)
{
  printf ("\n");
    
  unsigned int static_mem_top = platform->sram.offset + data_seg_size + bss_seg_size;
  int stack_bottom = (int)init_stack_ptr - depth;
  int middle = (int) stack_bottom - (int) static_mem_top;
  
  assert (stack_bottom > 0);
  
  printf ("memory map summary:\n");
  printf ("  data+BSS is at addresses 0x%x-0x%x\n",
	  platform->sram.offset,
	  static_mem_top);
  printf ("  stack (potentially) occupies addresses 0x%x-0x%x\n",
	  stack_bottom, 
	  init_stack_ptr);
  printf ("  there are 0x%x (%d) bytes left in the middle (for heap or whatever)\n", 
	  middle, middle);
  if (middle >= 0) {
    printf ("  system seems safe!\n");
  } else {
    printf ("  oops -- system is probably unsafe!\n");
  }
  printf ("\n");
}

static void find_duplicates (void)
{
  /*
   * hack -- treat interrupt vectors that point to the same address as
   * the same, this makes things more efficient and output clearer
   */
  for (unsigned int i=0; i<platform->num_vectors; i++) {
    int_info[i].duplicate = false;
  }  
  for (unsigned int i=0; i<platform->num_vectors; i++) {
    if (int_info[i].duplicate) continue;
    code_addr_t addr1 = platform->int_vecs[i].start_addr;
    code_addr_t target1 = all_code[addr1]->decode.target;
    for (unsigned int j=i+1; j<platform->num_vectors; j++) {
      code_addr_t addr2 = platform->int_vecs[j].start_addr;
      code_addr_t target2 = all_code[addr2]->decode.target;
      if (target1 == target2) {
	int_info[j].duplicate = true;
      }
    }
  }
}

static int total = 0;
static int steps = 0;
static int cnt = 0;

static void analyze_vector (int vec)
{
  int depth = 0;

  struct stack_queue_elt q1 = {
    ST_DEL_FLOW, 
    abs_loc(platform->int_vecs[vec].start_addr, (code_ctxt_t)0),
    abs_loc(0, (code_ctxt_t)0),
    0,
    0,
    (slot_t)0,
    zero_instance_cnt,
  };
  (void) enqueue_stack (&q1);

  struct stack_queue_elt q2 = {
    ST_ADD_FLOW, 
    abs_loc(platform->int_vecs[vec].start_addr, (code_ctxt_t)0),
    abs_loc(NO_CODE_ADDR, NO_CTX),
    0,                            // depth
    0,                            // path len
    NO_SLOT,                      // prev slot
    zero_instance_cnt,
  };
  (void) enqueue_stack (&q2);
  
  steps += drain_stack_queue (&depth);
  
  collect_results (vec);
  
  if (!int_info[vec].duplicate) {
    
    code_addr_t vector = platform->int_vecs[vec].start_addr;
    total += depth;
    cnt++;
    if (!quiet) {
      printf ("found an interrupt __vector_%d (%s) = %d, at %x\n", 
	      vector / VECTOR_SIZE,
	      platform->int_vecs[vec].name,
	      depth, 
	      wc_depth_loc[vec].addr);
    }
  }
}

static void report_deep_paths( int threshold ) {
  for (code_addr_t addr=0; addr < (code_addr_t) code_size; addr++) {
    if (all_code[addr] == NULL) continue;
    if (all_code[addr]->wc_path_depth < threshold) continue;

    if (avr_is_call(addr) && all_code[addr]->decode.target_set) {
      cout << "Deep edge" << "(" << dec << all_code[addr]->wc_path_depth << hex << ") "
           << addr_to_head_label(all_code[addr]->decode.target)
           << " : " 
           << addr_to_head_label(addr) 
           << endl;
    }
  }

}

void stack_scan (void)
{
  find_duplicates ();

  setup_stack_scan ();

  if (!quiet) {
    printf ("\n");
    printf ("worst-case stack depth requirements for individual entry points,\n");
    printf ("ignoring the possibility of self-stacking interrupts:\n");

    for (unsigned int vec=0; vec < platform->num_vectors; vec++) {
      if (indirectly_enabled[vec]) {
        cout << "WARNING: Treating irq " << vec << " as globally enabled" << endl;
      }
    }
  }

  for (unsigned int vec=0; vec < platform->num_vectors; vec++) {
    analyze_vector (vec);
  }

  if (!quiet) printf ("%d nontrivial interrupt vectors\n", cnt);
  
  if (mark_deepest_path) {
    report_deep_paths( 0 );
  }

  int depth = global_depth ();

  if (!quiet) {
    check_safety (depth);
    printf ("stack analysis steps = %d\n", steps);
  }

  if (!quiet || print_sum_depth) {
    printf ("\n");
    printf ("total stack requirement from naive summation = %d\n", total);
  }

  if (!quiet || print_whole_depth) {
    printf ("\n");
    printf ("total stack requirement from global analysis = %d\n", depth);
    printf ("\n");
  }
}

