#include "stacktool.h"

// misc globals
bool verbose = false;
int debug_data = 0;
int debug_depth = 0;
int debug_control = 0;
int debug_wcet = 0;
bool debug_elf = false;
bool do_wcet = false;
int print_max_path = 0;
bool print_disasm = false;
bool print_callgraph = false;
bool quiet = false;
bool print_sum_depth = false;
bool print_whole_depth = false;
bool mark_deepest_path = false;
const char *disasm_fn = 0;
unsigned int origin = 0;
const struct avr_platform *platform = NULL;
struct interrupt_info int_info[MAX_VECTORS+1];

static void usage (void)
{
  printf ("usage: stacktool [options] avr_elf_binary\n");
  printf ("  -c        make callgraphs that graphviz can render\n");
  printf ("  -d        dump disassembled program\n");
  printf ("  -D fn     dump disassembled program to file 'fn'\n");
  printf ("  -fcontrol extra verbosity about control flow analysis\n");
  printf ("  -fdata    extra verbosity about dataflow analysis\n");
  printf ("  -fdepth   extra verbosity about depth analysis\n");
  printf ("  -felf     extra verbosity about reading elf object files\n");
  printf ("  -fwcet    extra verbosity about wcet analysis\n");
  printf ("  -g        print interrupt preemption graph in dot format\n");
  printf ("  -m        mark deepest path reachable through each function\n");
  printf ("  -mmcu=xx  specify microcontroller type using gcc format (mandatory)\n");
  printf ("  -o addr   set origin of analysis to hex value addr\n");
  printf ("  -p        print path to maximum stack depth\n");
  printf ("  -reentrant x y -- permit interrupt x to reenter y times (default 1)\n");
  printf ("  -s        print only summation of individual entry-point stack depths\n");
  printf ("  -v        verbose information about abstract machine state\n");
  printf ("  -w        print only whole-program stack depth\n");
  printf ("  -wcet     do a wcet analysis\n");
  exit (EXIT_FAILURE);
}

static void hex_err (int pos)
{
  printf ("oops: hex number expected at cmd line position %d\n", pos);
  exit (-1);
}
  
static int expect_hex (const int argc, 
		       const char *argv[],
		       int pos)
{
  int i;
  char *c;
  const char *str = argv[pos];

  if (pos >= argc || str[0] == '\0') {
    hex_err (pos);
  }

  if (strcmp (str, "0x") == 0) {
    str += 2;
  }

  i = strtol (str, &c, 16);
  if (*c != '\0') {
    hex_err (pos);
  }

  return i;
}

static void int_err (int pos)
{
  printf ("oops: integer expected at cmd line position %d\n", pos);
  exit (-1);
}
  
static int expect_int (const int argc, 
		       const char *argv[], 
		       int pos)
{
  int i;
  char *c;
  const char *str = argv[pos];

  if (pos >= argc || str[0] == '\0') {
    int_err (pos);
  }

  i = strtol (str, &c, 10);
  if (*c != '\0') {
    int_err (pos);
  }

  return i;
}

const char *parse_args (const int argc, 
			const char *argv[])
{
  int arg = 0;

  /*
   * first arg that doesn't match the -x pattern has to be the input
   * file, and has to be the last argument in the list 
   */
  while (true) {

    arg++;
    if (arg+1 > argc) usage ();

    if (suffix_match(argv[arg],".cnt")) {
      read_wcet_from_file(argv[arg]);
      continue;
    }

    if (argv[arg][0] != '-') break;

    if (strcmp (argv[arg], "-c") == 0) {
      print_callgraph = true;
      continue;
    }

    if (strcmp (argv[arg], "-d") == 0) {
      print_disasm = true;
      continue;
    }

    if (strcmp (argv[arg], "-D") == 0) {
      print_disasm = true;
      arg++;
      if (arg+1 > argc) usage ();
      disasm_fn = argv[arg];
      continue;
    }

    if (strcmp (argv[arg], "-fcontrol") == 0) {
      debug_control++;
      continue;
    }
    if (strcmp (argv[arg], "-fdata") == 0) {
      debug_data++;
      continue;
    }
    if (strcmp (argv[arg], "-fdepth") == 0) {
      debug_depth++;
      continue;
    }
    if (strcmp (argv[arg], "-felf") == 0) {
      debug_elf = true;
      continue;
    }
    if (strcmp (argv[arg], "-fwcet") == 0) {
      debug_wcet++;
      continue;
    }

    if (strcmp (argv[arg], "-g") == 0) {
      print_ipg = true;
      continue;
    }

    if (strcmp (argv[arg], "-m") == 0) {
      mark_deepest_path = true;
      continue;
    }

    if (strncmp (argv[arg], "-mmcu=", 6) == 0) {
      int i;
      int found = 0;
      for (i=0; i<NUM_PLATFORMS; i++) {
	if (strcmp (argv[arg]+6, all_platforms[i]->name) == 0) {
	  platform = all_platforms[i];
	  found = 1;
	}
      }
      if (found == 0) {
	printf ("hmm... unknown microcontroller '%s'\n",
		argv[arg]+6);
	printf ("known chips are: ");
	for (i=0; i<NUM_PLATFORMS; i++) {
	  printf ("%s ", all_platforms[i]->name);	  
	}
	printf ("\n");
	die ();
      }
      continue;
    } 

    if (strcmp (argv[arg], "-o") == 0) {
      arg++;
      origin = expect_hex (argc, argv, arg);
      continue;
    }

    if (strcmp (argv[arg], "-p") == 0) {
      print_max_path++;
      continue;
    }

    if (strcmp (argv[arg], "-reentrant") == 0) {
      unsigned int vector;
      int count;
      if (platform == NULL) {
	printf ("oops -- reentrancy counts need to be specified after CPU type\n");
	die ();
      }
      arg++;
      vector = expect_int (argc, argv, arg);
      if (vector < 1 || vector >= platform->num_vectors) {
	printf ("oops -- vector %d out of range\n", vector);
	die ();
      }
      arg++;
      count = expect_int (argc, argv, arg);
      if (count < 1) {
	printf ("oops -- reentrancy count must be a natural number\n");
	die ();
      }
      int_info[vector].max_reenter = count;
    }

    if (strcmp (argv[arg], "-s") == 0) {
      quiet = true;
      print_sum_depth = true;
      continue;
    }

    if (strcmp (argv[arg], "-v") == 0) {
      verbose = true;
      continue;
    }

    if (strcmp (argv[arg], "-w") == 0) {
      quiet = true;
      print_whole_depth = true;
      continue;
    }

    if (strcmp (argv[arg], "-wcet") == 0) {
      do_wcet = true;
      continue;
    }

    usage ();
  }

  if (arg+1 != argc) usage ();

  return argv[arg];
}

